package webppclient;

import java.net.*;
import java.io.*;
import java.util.*;


///////////////////////////////////////////////////////////////////////////////
// 
// Latency table
//
///////////////////////////////////////////////////////////////////////////////
final class LatTable extends Hashtable implements Serializable
{
    // get best server from a list of servers =================================
    //    servers are looked up in this order:
    //      1) non-expired entries in latency percentile order
    //      2) expired entries in latency order
    //      3) new entries in random order
    //
    synchronized String getBestServer( Hashtable serverList )
        throws Exception
    {
        String serv;
        LatEntry entry;
        Vector newServ = new Vector();
        Vector valServ = new Vector();
        Vector expServ = new Vector();
        Date now       = new Date();

        for ( Enumeration e = serverList.keys();
              e.hasMoreElements(); ) {
	        
            serv = (String) e.nextElement();
            entry = (LatEntry) get( serv );
	
            if ( entry == null ) {
	            // new server, create an entry
                entry = new LatEntry(Webpp.STAT_SIZE);
                put( serv, entry );
            }

	    //entry.up = true;	// Mehmet: added for resolving a problem. Normally this shouldn't be used at all!!!
	
	    if ( entry.up ) {	// server is reachable
	            
                if ( entry.ts == null ) {
	                // add to a list of new servers
                    newServ.addElement( serv );
	
                } else if ( now.getTime() - entry.ts.getTime() < Webpp.TTL ) {
	                // add to alist of valid servers
                    valServ.addElement( serv );
	
                } else {
	                // add to a list of expired servers
                    expServ.addElement( serv );
                }
                
            } else {
                // if server is unreachable then add it to expired servers
                // after a long timeout
                
                if ( now.getTime() - entry.ts.getTime() > entry.failPeriod ) {
                    // add to a list of new servers (to guarantee an 
                    // immediate sample)
                    getAsynchSample( serv );
                }
            }
        }

        if( Webpp.SERVER_CHOICE_DEBUG ){
            int i;
            
            System.out.print( "non-expired servers: " );
            for( i = 0; i < valServ.size(); i++ ){
                System.out.print( valServ.elementAt(i) + " " );
            }
            System.out.println( " " );
            
            System.out.print( "expired servers: " );
            for( i = 0; i < expServ.size(); i++ ){
                System.out.print( expServ.elementAt(i) + " " );
            }
            System.out.println( " " );
            
            System.out.print( "new servers: " );
            for( i = 0; i < newServ.size(); i++ ){
                System.out.print( newServ.elementAt(i) + " " );
            }
            System.out.println( " " );
        }
            
            
        if ( valServ.size() > 0 ) {
	    // try to get fastest non-expired server
             
	    if ( Webpp.SERVER_CHOICE_DEBUG ) {
                System.out.println( "*** non-expired ***" );
            }

            // new servers have prio over expired servers (new servers are
            // viewed as servers with maximally old timestamps)
            if ( newServ.size() > 0 ) {
                getAsynchSample( getRandomServer( newServ ) );
    
            } else if ( expServ.size() > 0 ) {
                getAsynchSample( getOldestServer( expServ ) );
            }
	
            return getMinLatServer( valServ );
	
        } else if ( expServ.size() > 0 ) {
            // otherwise, try to get fastest expired server
	
	    if ( Webpp.SERVER_CHOICE_DEBUG ) {
                System.out.println( "*** expired ***" );
            }
             
            if ( newServ.size() > 0 ) {
                getAsynchSample( getRandomServer( newServ ) );
            }
	
            return getMinLatServer( expServ );
        
        } else if ( newServ.size() > 0 ) {
	    // otherwise, pick a random new server
	        
	    if ( Webpp.SERVER_CHOICE_DEBUG ) {
                System.out.println( "*** random ***" );
            }
	        
            return getRandomServer( newServ );
             
        } else {
            // no choices left - all servers are down
            return null;
        }
    }

    // Mehmet: insert a new latency table item for a server
    synchronized void insertLatency( String server, double lat )
    {
        LatEntry entry = new LatEntry(Webpp.STAT_SIZE);
        Assert.assert( entry != null, server );
        put( server, entry );

        // add new latency
        entry.lat.addDouble( lat );
        
        // new timestamp for the entry
        entry.ts = new Date();
    }

    // add new latency sample ================================================
    synchronized void addLatency( String server, double lat )
    {
        LatEntry entry = (LatEntry) get( server );
        Assert.assert( entry != null, server );

        // add new latency
        entry.lat.addDouble( lat );
        
        // new timestamp for the entry
        entry.ts = new Date();
    }

    // indicate a pending sample ===========================================
    synchronized void sentSample( String server, boolean val )
    {
        LatEntry entry = (LatEntry) get( server );
        Assert.assert( entry != null, server );

        // add new latency
        entry.sentSample = val;
    }

    // query if a pending sample ===========================================
    synchronized boolean sentSample( String server )
    {
        LatEntry entry = (LatEntry) get( server );
        Assert.assert( entry != null, server );

        // add new latency
        return entry.sentSample;
    }

    // update reachability status ============================================= 
    synchronized void updateReachability( String server, boolean status )
    {
        LatEntry entry = (LatEntry) get( server );
        Assert.assert( entry != null );
	
        // update status
        entry.up = status;
        
        // new timestamp for the entry
        entry.ts = new Date();
    }

    // double fail persiod for server =====================================
    synchronized void incrFailPeriod( String server )
    {
        LatEntry entry = (LatEntry) get( server );
        Assert.assert( entry != null );
        
        // double the fail period
        entry.failPeriod *= 2;
        
        // new timestamp for the entry
        entry.ts = new Date();
    }
    
    // reset fail persiod for server =======================================
    synchronized void resetFailPeriod( String server )
    {
        LatEntry entry = (LatEntry) get( server );
        Assert.assert( entry != null );
        
        // double the fail period
        entry.failPeriod = Webpp.FAIL_PERIOD_INI;
        
        // new timestamp for the entry
        entry.ts = new Date();
    }
    
    // print contents of latency table ========================================
    synchronized void print()
    {
        String serv;
        LatEntry entry;

        for ( Enumeration k = keys();
              k.hasMoreElements(); ) {

            serv = (String) k.nextElement();
            entry = (LatEntry) get( serv );

            System.out.print( serv + ":: " );
            entry.print();
            System.out.println();
        }    
    }

    // send asynch a request for bogus document and time the latency ==========
    void getAsynchSample( String server )
    {
        if( Webpp.SERVER_CHOICE_DEBUG ){
            System.out.println( "starting asynch sample for " + server );
        }
        
        // start the thread
        new AsynchSampleThread( server ).start();
    }

    // get server with minimal latency percentile =============================
    private String getMinLatServer( Vector serverList )
        throws Exception
    {
        double minLat = Double.MAX_VALUE;
        double tmpLat;
        LatEntry entry;
        String serv;
        String minServ = null;
        int sz;

        for ( Enumeration e = serverList.elements();
              e.hasMoreElements(); ) {   

            serv = (String) e.nextElement();
            entry = (LatEntry) get( serv );
            
            // ignore servers with no samples (this is a safeguard)
            if( entry.lat.size() <= 0 ){
                continue;
            }
            
            tmpLat = entry.lat.getStatMoments().
                     getPercentile(Webpp.PERCENTILE, entry.lat.size());

            if ( minLat > tmpLat ) {
                 minLat  = tmpLat;
                 minServ = serv;
            }
        }
        
        Assert.assert( minServ != null );
        return minServ;
    }

    // get server with minimal latency percentile =============================
    private String getRandomServer( Vector serverList )
    {
        return (String) 
                 serverList.elementAt( Math.abs(Net.rnd.nextInt()) % 
                                       serverList.size() );
    }

    // get server with oldest timestamp =======================================
    private String getOldestServer( Vector serverList )
    {
        String serv;
        String oldestServ = null;
        LatEntry entry;
        long currTs;
        long oldestTs = Long.MAX_VALUE;

        Assert.assert( serverList.size() > 0 );

        for ( Enumeration e = serverList.elements();
              e.hasMoreElements(); ) {   

            serv = (String) e.nextElement();
            entry = (LatEntry) get( serv );

            // the passed list of servers should contain timestamps
            Assert.assert( entry.ts != null );

            if ( oldestTs > (currTs = entry.ts.getTime()) ) {
                oldestTs = currTs;
                oldestServ = serv;
            }
        }

        // must find at least one server since the list is non-empty
        Assert.assert( oldestServ != null );

        return oldestServ;
    }
}
    


///////////////////////////////////////////////////////////////////////////////
// 
// Hash entry for latency table
//
///////////////////////////////////////////////////////////////////////////////
final class LatEntry implements Serializable
{
    // constructor ============================================================
    LatEntry( int size )
    {
        Assert.assert( size > 0, Integer.toString(size) );

        lat        = new BoundStatVector( size );
        ts         = null;
        up         = true;
        failPeriod = Webpp.FAIL_PERIOD_INI;
        sentSample = false;
    }

    // print latency table entry ==============================================
    void print()
    {
        System.out.print( "Timestamp: " + ts + "   " );
        System.out.print( "Reachable: " + up + "   " );
        System.out.print( "failPeriod: " + failPeriod + "   " );
        System.out.print( "sentSample: " + sentSample + "   " );
        lat.print();
    }

    // disable default constructor ============================================
    private LatEntry()
    {
        ;
    }
    
    // data members ===========================================================
    Date ts;
    boolean up;
    BoundStatVector lat;
    long failPeriod;
    boolean sentSample;
}


///////////////////////////////////////////////////////////////////////////////
// 
// Thread for asynchronous sampling
//
///////////////////////////////////////////////////////////////////////////////
final class AsynchSampleThread extends Thread
{
    // constructor ============================================================
    AsynchSampleThread( String s )
    {
        // initialize the server to be sampled
        server = s;
    }
    
    // sends a request for non-existing document ==============================
    //    (in the future may be replaced by a HEAD request)
    //
    public void run()
    {
        // give up if there is already a pending sample
        if( Webpp.lat.sentSample( server ) ){
            if( Webpp.SERVER_CHOICE_DEBUG ){
                System.out.println( "giving up sampling at " + server );
            }
            return;
        } else {
            Webpp.lat.sentSample( server, true );
        }
        
        // send a sample
        String url = "http://" + server + "/dummy.html";
	System.out.println("Async sample for : " + url);
        Stat1 stat = Net.getSample( url );

	// Mehmet: get two samples and pick the minimum
	double myLatency = 0;
	if( stat != null ) {
	    myLatency = stat.lat;
	    stat = Net.getSample( url );
	}

        // process response
        if( stat != null ) {
	    if (myLatency == 0 || myLatency > stat.lat) myLatency = stat.lat;	// Mehmet: pick smaller latency
	    Webpp.lat.addLatency( server, myLatency );
            Webpp.lat.updateReachability( server, true );
            Webpp.lat.resetFailPeriod( server );
            System.out.println("Async sample time was: " + myLatency );
	    Webpp.writeToLogWithTime("AsyncSample " + server + " " + Double.toString(myLatency * 1000));
            if( Webpp.SERVER_CHOICE_DEBUG ){
                System.out.println( "Got asynch sample for " + 
                                    server + " value: " + 
                                    Double.toString( myLatency ) );
            }
        } else {
	    System.out.println("Async sample was unsuccessful");
	    Webpp.writeToLogWithTime("Unreachable " + server + " AsyncSample");
            Webpp.lat.updateReachability( server, false );
            Webpp.lat.incrFailPeriod( server );
            
            if( Webpp.SERVER_CHOICE_DEBUG ){
                System.out.println( "No asynch sample from " + 
                                    server + " doubling fail period" );
            }            
        } 
        
        Webpp.lat.sentSample( server, false );
    }

    // private data members ===================================================
    private String server;
}
