package webppclient;

// Web++ Client --- Response Times Table
// RespTimeTable class: keeps track of server response times

import java.util.*;
import java.lang.*;
import java.io.*;

final class RespTimeTable extends Hashtable implements Serializable
{

  RespTimeTable()
  { super(); }

  // inTable returns true or false depending on whether a given
  // URL's response time is in the table or not.
  synchronized boolean inTable(String url)
  {
    if (containsKey(url))
    {
      //System.out.println("In RespTimeTable: " + url);
      return true;
    }
    else
    {
      //System.out.println("Not in RespTimeTable: " + url);
      return false;
    }
  }

  // putHostName inserts the closest (host) server's name into the table
  synchronized void putHostName(String hname)
  {
    long ftime = System.currentTimeMillis();
    //System.out.println("Putting host server name as: " + hname);
    RespTimeTableItem item = new RespTimeTableItem("HostServer", hname, 0, ftime);
    put("HostServer", item);
  }

  // getHostName returns the closest (host) server's name if there is one.
  synchronized String getHostName()
  {
    RespTimeTableItem item = (RespTimeTableItem) get("HostServer");
    if (item != null) return item.docName;
    else return null;
  }

  // putRT inserts the given URL's response time to the response times table
  // with the current time as the last update time.
  synchronized void putRT(String url, String docname, long rt)
  {
    long ftime = System.currentTimeMillis();
    //System.out.println("Putting response time for: " + url);
    RespTimeTableItem item = new RespTimeTableItem(url, docname, rt, ftime);
    put(url, item);
  }

  // getDocnameAndRT returns document name and the last response time of the
  // request that was previously sent to the given URL.
  synchronized String getDocnameAndRT(String url)
  {
    RespTimeTableItem item = (RespTimeTableItem) get(url);
    if (item != null)
      return (item.docName + "|" + (new Long(item.respTime)).toString());
    else return null;
  }

  // getRT returns the last response time of the request that was previously
  // sent to the given URL.
  synchronized long getRT(String url)
  {
    RespTimeTableItem item = (RespTimeTableItem) get(url);
    if (item != null) return item.respTime;
    else return 0;
  }

  // getDocname returns the document name of the request that was previously
  // sent to the given URL.
  synchronized String getDocname(String url)
  {
    RespTimeTableItem item = (RespTimeTableItem) get(url);
    if (item != null) return item.docName;
    else return null;
  }

}


// =============================================================
// RespTimeTableItem class: used in response time table for the
// 'value' field objects of <key,value> pairs in the table.
// =============================================================

final class RespTimeTableItem implements Serializable
{
  String url;		// server URL
  String docName;	// document name of the last request to that server
  long respTime;	// response time for that document
  long lastUpdateTime;	// last update time of the table item for this URL

  public RespTimeTableItem(String fn, String doc, long rt, long ft)
  {
    url = fn;
    docName = doc;
    respTime = rt;
    lastUpdateTime = ft;
  }

  private RespTimeTableItem()
  { ; }
}
